// VOIDSTAR - Goofing around, Sept. 2024 (approach borrowed from DestinyHunter)
#include <core.h>
#include <divmod4_table.h>

#define MAX_MAP_ROWS 100
#define MAX_MAP_TOP 10
#define MAX_MAP_BOTTOM 90

#define MAX_MAP_COLS 252
#define MAX_MAP_COLSML 20
#define MAX_MAP_COLSMR 232

static char g_main_map[MAX_MAP_ROWS][MAX_MAP_COLS];

static unsigned char or_equal_modifier[] = {0xC0,0x30,0x0C,0x03};

static unsigned char cell_state[MAX_MAP_ROWS][63];  // Y/X, 63*(4 half-nibbles per cell) = 252 columns

typedef struct 
{
  unsigned char offset_x;
  unsigned char offset_y;
  unsigned char color;
} Location_to_draw;

#define MAX_LOCATIONS_TO_DRAW 300
static Location_to_draw locations_to_draw[MAX_LOCATIONS_TO_DRAW];

static unsigned char num_locations_to_draw = 0;	

void write_pixel(unsigned char x, unsigned char y, unsigned char col)
{
    unsigned char n1,n2,n3;
	unsigned long ofs;
	
	ofs = (y*320LU) + x;
	
    n1 = (ofs & 0xFF);
    n2 = (ofs >> 8) & 0xFF;
    n3 = (ofs >> 16) & 0x01;
	
	POKE(0x9F20,n1);
    POKE(0x9F21,n2);
    POKE(0x9F22,n3);
	
	POKE(0x9F23,col);
}

void write_pixel_fast(unsigned long ofs, unsigned char col)
{
    unsigned char n1,n2,n3;	
	
	n1 = (ofs & 0xFF);
    n2 = (ofs >> 8) & 0xFF;
    n3 = (ofs >> 16) & 0x01;
	
	POKE(0x9F20,n1);
    POKE(0x9F21,n2);
    POKE(0x9F22,n3);
	
	POKE(0x9F23,col);
}

void write_pixel_incI(unsigned long ofs, unsigned char col)
{
    unsigned char n1,n2,n3;	
	
	n1 = (ofs & 0xFF);
    n2 = (ofs >> 8) & 0xFF;
    n3 = (ofs >> 16) & 0x01;  // enable auto increment 1 byte
	
	POKE(0x9F20,n1);
    POKE(0x9F21,n2);
    POKE(0x9F22,n3 | 0x10);
	
	POKE(0x9F23,col);
}

#define WRITE_PIXEL_INC(c) \
  POKE(0x9F23,c);

#define MARK_LOCATION_AS_DRAWN(data_x,data_y)    \
	cell_state[ data_y ][ div4_table[data_x] ] |= or_equal_modifier[mod4_table[data_x]];	

void BUFFER_LOCATION_TO_DRAW(unsigned char data_x, unsigned char data_y, unsigned char a_color)
{
	MARK_LOCATION_AS_DRAWN(data_x, data_y);
	locations_to_draw[num_locations_to_draw].offset_x = data_x;
	locations_to_draw[num_locations_to_draw].offset_y = data_y;
	locations_to_draw[num_locations_to_draw].color = a_color;
	++num_locations_to_draw;
}

void run_stage()
{
	Location_to_draw* location_to_draw_ptr;	
	
	register unsigned char i;  //< Since this is used in a lot of logic, we keep the name short.  But it should be called: temp_cell_state
	register unsigned char temp_x;  //< temp_x used for walking the cell_state, also used to move icons, and in drawing the icons (just the portion that would fit on the screen)
	register unsigned char temp_y;  //< temp_y used for walking the cell_state, also used to move icons, and in drawing the icons (just the portion that would fit on the screen)

	unsigned char* ptr_map;	
	unsigned char* ptr_cell;	
	unsigned int x_addr;
	
	// ** RESTORE STAGE MAP PORTIONS THAT ARE NO LONGER COVERED ******** 
	{			
		ptr_map = (unsigned char*)(&g_main_map[0][0]);
		ptr_cell = (unsigned char*)(&cell_state[0][0]);			
		
		temp_x = 0;  // keep track of where we are currently processing (horizontal)
		temp_y = 0;  // keep track of where we are currently processing (vertical)

		x_addr = cx16_screen_row_offset[temp_y];

		while (TRUE)  
		{			
			{			
				if (temp_x >= MAX_MAP_COLS)
				{						
					++temp_y;
					if (temp_y >= MAX_MAP_ROWS) break;
					
					x_addr = cx16_screen_row_offset[temp_y];
					temp_x = 0;
				}					
					
				// 0x11 (3) was drawn this last animation frame (already on the screen)
				//      (2) could be used to let icons linger one extra frame (not used here)
				// 0x01 (1) was previously drawn on -- animation may stay (increase it back to 2) or it will drop back 0 (background)
				// 0x00 (0) no change
					
				i = (*ptr_cell);
					
				// 1 2 3 4 5 6 7 8
				// [1] [2] [3] [4]
										
				if (i == 0)
				{
					// nothing to do, no animation impacts for this group, fast-track to next
					x_addr += 4;
					ptr_map += 4;
					
					temp_x += 4;
				}
				else
				{					
					// 7 6 5 4 3 2 1 0
					// 1 1 0 0 0 0 0 0 			C0
					// 0 0 1 1 1 1 1 1          3F
					// 0 1 0 0 0 0 0 0          40
					if ((i & 0xC0) == 0xC0) { i &= 0x3F; i |= 0x40; }  // reduce cell_1_state from 2 to 1
					else if ((i & 0xC0) == 0x40)
					{
						// draw map at cell_1 location
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0x3F;  //< reduce cell_1_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;
					
					// 7 6 5 4 3 2 1 0
					// 0 0 1 1 0 0 0 0 			30
					// 1 1 0 0 1 1 1 1          CF
					// 0 0 0 1 0 0 0 0          10						
					if ((i & 0x30) == 0x30) { i &= 0xCF; i |= 0x10; }  // reduce cell_2_state from 2 to 1
					else if ((i & 0x30) == 0x10)
					{
						// draw map at cell_2 location
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0xCF;  //< reduce cell_1_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;
						
					// 7 6 5 4 3 2 1 0
					// 0 0 0 0 1 1 0 0 			0C
					// 1 1 1 1 0 0 1 1          F3
					// 0 0 0 0 0 1 0 0          04
					if ((i & 0x0C) == 0x0C) { i &= 0xF3; i |= 0x04; }  // reduce cell_2_state from 2 to 1
					else if ((i & 0x0C) == 0x04)
					{
						// draw map at cell_3 location, 
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0xF3;  //< reduce cell_3_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;
					
					// 7 6 5 4 3 2 1 0
					// 0 0 0 0 0 0 1 1 			03
					// 1 1 1 1 1 1 0 0          FC
					// 0 0 0 0 0 0 0 1          01						
					if ((i & 0x03) == 0x03) { i &= 0xFC; i |= 0x01; }  // reduce cell_2_state from 2 to 1
					else if ((i & 0x03) == 0x01)
					{
						// draw map at cell_4 location
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0xFC;  //< reduce cell_4_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;

					(*ptr_cell) = i;
				}
					
				++ptr_cell;				
			}
		}
		// ******************************************

		// ** DRAW ALL CELLS MARKED TO BE DRAWN **************************		
	    location_to_draw_ptr =  &locations_to_draw[num_locations_to_draw-1];
		while (num_locations_to_draw > 0)
		{		    
			write_pixel(
				location_to_draw_ptr->offset_x,
				location_to_draw_ptr->offset_y,
				location_to_draw_ptr->color
			);	
						
			--num_locations_to_draw;
			--location_to_draw_ptr;
		}
		// ***************************************************************
	}	
}

void enable_screen_80()
{
	__asm__("lda #$80");  // 320x240
	__asm__("clc");
	__asm__("jsr $ff5f");
}

void show_mouse()
{
	__asm__("sec");
	__asm__("jsr $ff5f");  // get current screen mode
	__asm__("lda #1");     
	__asm__("jsr $ff68");  // show default mouse cursor
}

void poll_mouse()
{
	__asm__("LDX #$70");
    __asm__("JSR $ff6b");  // get mouse position in $70/$71 (X) and $72/$73 (Y)
	__asm__("sta $400");   // store button-state into golden RAM
}

void clear_grx_screen(unsigned char color)
{
  unsigned int ofs = 4800U;  // proportional to the number of inline POKE's done below
  
  // enable auto-byte increment, init offset to VERA addr 0
  POKE(0x9F22,0x10);
  POKE(0x9F21,0);
  POKE(0x9F20,0);
    
  while (ofs > 0U)
  {
    POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
    POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	POKE(0x9F23, color);
	--ofs;
  }  
  
  // turn off auto-increment
  POKE(0x9F22, 0);
}

     //                           0  1   2  3  4  5  6
static unsigned char col_seq[] = {4, 14, 3, 5, 7, 8, 2};  // X16 butterfly colors sequence

int main()
{
  unsigned int x,tx;
  unsigned int y,ty;
  unsigned int cx, cy, delta, ax, ay, dir;
  unsigned char c,cstart;
  unsigned long ofs, i;
  unsigned char mbutt;
  unsigned char ch;
   
  enable_screen_80();
  
  show_mouse();
  
  clear_grx_screen(0);
  
  // ===============================
  // generate a background (emulate carbon fiber) - this is slow in code, but would be replaced
  // by a BVLOAD in final
  c = 20;   // start at some gray color
  for (y = 0; y < MAX_MAP_ROWS; ++y)
  {
	  ofs = cx16_screen_row_offset[y];
	  write_pixel_incI(ofs, c);  ++c;
	  
	  for (x = 1; x < MAX_MAP_COLS; ++x)
	  {
		  g_main_map[y][x] = c;
		  		  
		  WRITE_PIXEL_INC(c);
		  
		  ++c;
		  if (c > 22) c = 20;  // cycle through some grays
	  }
  }
  // ===============================
  
  // specify an initial object center x/y to start drawing at
  cx = 126;
  cy = 50;
  
  cstart = 0;    // color index to start at
  dir = 0;  // color-spin direction
  delta = 1;  // how much delta to apply this cycle
  i = 0;  // main loop counter
  while (TRUE)
  {
	++i;
	
	GET_CX16_KEY(ch);
	
	// arrow key movement
	if (ch == KEY_RIGHT_ARROW) ++cx;
	else if (ch == KEY_LEFT_ARROW) --cx;
	else if (ch == KEY_DOWN_ARROW) ++cy;
	else if (ch == KEY_UP_ARROW) --cy;
	
	// ==== bounce into margin walls =======
	if (cx > MAX_MAP_COLSMR) cx = MAX_MAP_COLSMR;
	else if (cx < MAX_MAP_COLSML) cx = MAX_MAP_COLSML;
	if (cy > MAX_MAP_BOTTOM) cy = MAX_MAP_BOTTOM;
	else if (cy < MAX_MAP_TOP) cy = MAX_MAP_TOP;
	// ====================================
	
	if (i % 4 == 0)  // optional; don't really need to run the main loop all the time (replace with a jiffies-timer)
	{
		run_stage();
		
		c = cstart;
		
		if (dir == 0) ay=cy-3;
		else ay=cy+4-1;
		
		for (y = cy-3; y < cy+4; ++y)
		{
		  if (dir == 0) ax=cx-5;
		  else ax=cx+5-1;
		  
		  for (x = cx-5; x < cx+5; ++x)
		  {
			   BUFFER_LOCATION_TO_DRAW(ax, ay, col_seq[c]);  
			   if (dir == 0) ax += delta;
			   else ax -= delta;
		  }  
		  ++c;
		  if (c > 6) c = 0;
		  
		  if (dir == 0) ay += delta;
		  else ay -= delta;
		}
		
		if (i % 20 == 0)  // just a way to slow down color cycling without using timer
		{
			++delta;
			if (delta > 2) 
			{
				delta = 1;
				
				// switch direction
				if (dir == 0) dir = 1;
				else dir = 0;
			}
		}
		
		++cstart;  // offset over one to make the colors cycle around
		if (cstart > 6) cstart = 0;
	}
	
	poll_mouse();
	
	mbutt = PEEK(0x400);
	x = PEEKW(0x70);
	y = PEEKW(0x72);
	// if pressed any button and are inside the carbon fiber region...
	if ((mbutt != 0) && (x < MAX_MAP_COLS) && (y < MAX_MAP_ROWS))
	{		
		g_main_map[y][x] = 1;
	    
		if (
		  (x > MAX_MAP_COLSML) && (x < MAX_MAP_COLSMR)
		  && (y > MAX_MAP_TOP) && (y < MAX_MAP_BOTTOM)
		)
		{
			for (ty=y-mbutt; ty<(y+mbutt); ++ty)
			{
				BUFFER_LOCATION_TO_DRAW(x,ty, 5);
			}
		
			for (tx=x-mbutt; tx<(x+mbutt); ++tx)
			{
				BUFFER_LOCATION_TO_DRAW(tx,y, 5);
			}
			
		}
		else  // just do a single dot within the margin space
		{
			BUFFER_LOCATION_TO_DRAW(x,y, 5);
		}
			
	}
	
  }
  
  return 0;
}
